/***************************************************************************
 * Copyright 1998-2018 by authors (see AUTHORS.txt)                        *
 *                                                                         *
 *   This file is part of LuxCoreRender.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License");         *
 * you may not use this file except in compliance with the License.        *
 * You may obtain a copy of the License at                                 *
 *                                                                         *
 *     http://www.apache.org/licenses/LICENSE-2.0                          *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.*
 * See the License for the specific language governing permissions and     *
 * limitations under the License.                                          *
 ***************************************************************************/

#include <boost/format.hpp>

#include "slg/engines/tilerepository.h"

using namespace std;
using namespace luxrays;
using namespace slg;

//------------------------------------------------------------------------------
// TileCoord
//------------------------------------------------------------------------------

BOOST_CLASS_EXPORT_IMPLEMENT(slg::TileRepository::Tile::TileCoord)

template<class Archive> void TileRepository::Tile::TileCoord::serialize(Archive &ar, const unsigned int version) {
	ar & x;
	ar & y;
	ar & width;
	ar & height;
}

namespace slg {
// Explicit instantiations for portable archives
template void TileRepository::Tile::serialize(LuxOutputArchive &ar, const unsigned int version);
template void TileRepository::Tile::serialize(LuxInputArchive &ar, const unsigned int version);
// Explicit instantiations for polymorphic archives
template void TileRepository::Tile::serialize(boost::archive::polymorphic_oarchive &ar, const u_int version);
template void TileRepository::Tile::serialize(boost::archive::polymorphic_iarchive &ar, const u_int version);
}

//------------------------------------------------------------------------------
// Tile
//------------------------------------------------------------------------------

BOOST_CLASS_EXPORT_IMPLEMENT(slg::TileRepository::Tile)

template<class Archive> void TileRepository::Tile::load(Archive &ar, const u_int version) {
	ar & coord;
	ar & pass;
	ar & error;
	ar & done;

	// tileRepository has to be set from the code using the serialization
	tileRepository = NULL;

	ar & allPassFilm;
	// Disable OpenCL
	allPassFilm->oclEnable = false;

	ar & evenPassFilm;
	// Disable OpenCL
	evenPassFilm->oclEnable = false;

	ar & allPassFilmTotalYValue;
	ar & hasEnoughWarmUpSample;
}

template<class Archive> void TileRepository::Tile::save(Archive &ar, const u_int version) const {
	ar & coord;
	ar & pass;
	ar & error;
	ar & done;

	ar & allPassFilm;
	ar & evenPassFilm;
	ar & allPassFilmTotalYValue;
	ar & hasEnoughWarmUpSample;
}

namespace slg {
// Explicit instantiations for portable archives
template void TileRepository::Tile::save(LuxOutputArchive &ar, const u_int version) const;
template void TileRepository::Tile::load(LuxInputArchive &ar, const u_int version);
// Explicit instantiations for polymorphic archives
template void TileRepository::Tile::save(boost::archive::polymorphic_oarchive &ar, const u_int version) const;
template void TileRepository::Tile::load(boost::archive::polymorphic_iarchive &ar, const u_int version);
}

//------------------------------------------------------------------------------
// TileRepository
//------------------------------------------------------------------------------

BOOST_CLASS_EXPORT_IMPLEMENT(slg::TileRepository)

template<class Archive> void TileRepository::load(Archive &ar, const u_int version) {
	boost::unique_lock<boost::mutex> lock(tileMutex);

	ar & tileWidth;
	ar & tileHeight;
	ar & maxPassCount;
	ar & convergenceTestThreshold;
	ar & convergenceTestThresholdReduction;
	ar & convergenceTestWarmUpSamples;
	ar & varianceClamping;
	ar & enableMultipassRendering;
	ar & enableRenderingDonePrint;
	ar & done;

	ar & startTime;
	ar & filmRegionWidth;
	ar & filmRegionHeight;
	ar & filmTotalYValue;
	ar & tileList;

	u_int todoListSize;
	ar & todoListSize;
	for (u_int i = 0; i < todoListSize; ++i) {
		Tile *todoTile;
		ar & todoTile;
		todoTiles.push(todoTile);
	}

	pendingTiles.resize(0);
	ar & convergedTiles;

	// Initialize the Tile::tileRepository field
	BOOST_FOREACH(Tile *tile, tileList)
		tile->tileRepository = this;
}

template<class Archive> void TileRepository::save(Archive &ar, const u_int version) const {
	boost::unique_lock<boost::mutex> lock(tileMutex);

	ar & tileWidth;
	ar & tileHeight;
	ar & maxPassCount;
	ar & convergenceTestThreshold;
	ar & convergenceTestThresholdReduction;
	ar & convergenceTestWarmUpSamples;
	ar & varianceClamping;
	ar & enableMultipassRendering;
	ar & enableRenderingDonePrint;
	ar & done;

	ar & startTime;
	ar & filmRegionWidth;
	ar & filmRegionHeight;
	ar & filmTotalYValue;
	ar & tileList;

	const u_int count = todoTiles.size() + pendingTiles.size();
	ar & count;
	BOOST_FOREACH(Tile *tile, todoTiles)
		ar & tile;
	BOOST_FOREACH(Tile *tile, pendingTiles)
		ar & tile;

	ar & convergedTiles;
}

namespace slg {
// Explicit instantiations for portable archives
template void TileRepository::save(LuxOutputArchive &ar, const u_int version) const;
template void TileRepository::load(LuxInputArchive &ar, const unsigned int version);
// Explicit instantiations for polymorphic archives
template void TileRepository::save(boost::archive::polymorphic_oarchive &ar, const u_int version) const;
template void TileRepository::load(boost::archive::polymorphic_iarchive &ar, const u_int version);
}
